/*
Copyright (c) 2020 Steffen Illhardt

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
the Software, and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#if defined(_MSC_VER)
# define _CRT_SECURE_NO_WARNINGS // Microsoft considers sprintf to be unsafe
#endif
#include "conprogressbar.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#undef WINVER
#define WINVER 0x0601
#undef _WIN32_WINNT
#define _WIN32_WINNT 0x0601
#undef WIN32_LEAN_AND_MEAN
#define WIN32_LEAN_AND_MEAN
#include <windows.h>

struct tag_progress_bar
{
  CRITICAL_SECTION mtx; // structure used for mutual exclusion
  CONSOLE_CURSOR_INFO cursor_inf; // size of the cursor
  SMALL_RECT bar_target; // rect where the progress bar is intended to be written
  SMALL_RECT bar_metrics; // rect where the progress bar has been actually written
  HANDLE h_out; // handle referencing the console output buffer
  HANDLE h_thread; // thread handle
  char *msg; // message to be displayed
  size_t steps; // total number of steps
  size_t step_count; // current number of steps
  WORD color; // holds the swapped background and foreground colors to draw the progress
  short signal; // if not 0, force pbar_push returning false to any thread which is still trying to invoke it
  COORD bar_size; // size of the bar array
  CHAR_INFO bar[1]; // placeholder for as many CHAR_INFO elements as necessary for the length of the bar
};

static DWORD __stdcall write_bar(void *arg) // thread function
{
  progress_bar pbar = arg;
  const size_t bar_len = pbar->step_count * pbar->bar_size.X / pbar->steps;
  for (CHAR_INFO *it = pbar->bar + 0, *const end = it + bar_len; it < end; ++it)
    it->Attributes = pbar->color;

  const short half = pbar->bar_size.X >> 1;
  if (half > 1)
  {
    char percentage[4] = { 0 };
    sprintf(percentage, "%3u", (unsigned)(pbar->step_count * 100 / pbar->steps));
    (pbar->bar + half - 2)->Char.AsciiChar = percentage[0];
    (pbar->bar + half - 1)->Char.AsciiChar = percentage[1];
    (pbar->bar + half)->Char.AsciiChar = percentage[2];
  }

  pbar->bar_metrics = pbar->bar_target;
  WriteConsoleOutputW(pbar->h_out, pbar->bar, pbar->bar_size, (COORD){ 0 }, &pbar->bar_metrics);

  if (*pbar->msg != '\0')
  {
    DWORD processed = 0UL;
    WriteConsoleOutputCharacterA(pbar->h_out, pbar->msg, (DWORD)(pbar->bar_metrics.Right - pbar->bar_metrics.Left + 1), (COORD){ pbar->bar_metrics.Left, pbar->bar_metrics.Bottom + 1 }, &processed);
  }

  return 0UL;
}

progress_bar pbar_create(size_t steps, short x, short y, bool abs, short len, unsigned char col, const char *msg)
{
  if (!steps || !len)
    return NULL;

  // Explicitly use the console output buffer because the stdout might be redirected.
  HANDLE h_out = CreateFileW(L"CONOUT$", GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, 0UL, NULL);
  if (h_out == INVALID_HANDLE_VALUE)
    return NULL;

  CONSOLE_SCREEN_BUFFER_INFO csbi = { 0 };
  if (!GetConsoleScreenBufferInfo(h_out, &csbi) || x > csbi.dwSize.X || len > csbi.dwSize.X || y > csbi.dwSize.Y)
  {
    CloseHandle(h_out);
    return NULL;
  }

  // Update x, y, and len depending on their original values and on the characteristics of the current screen buffer.
  if (x < 0)
    x = 1;

  if (y < 0)
  {
    abs = true;
    y = csbi.dwCursorPosition.X == 0 ? csbi.dwCursorPosition.Y : csbi.dwCursorPosition.Y + 1;
  }

  if (!abs)
    y += csbi.srWindow.Top;

  if (len < 0 || (x + len) > (csbi.dwSize.X - 1))
    len = csbi.dwSize.X - x - 1;

  // Check whether the properties are reasonable. Allocate zero-initialized memory for the progress bar structure.
  progress_bar pbar = NULL;
  if (y < 0 || y + 3 > csbi.dwSize.Y || len < 1 || (pbar = calloc(1U, (offsetof(struct tag_progress_bar, bar)) + len * sizeof(CHAR_INFO))) == NULL)
  {
    CloseHandle(h_out);
    return NULL;
  }

  // Initialize text and color of every character cell in the progress bar.
  for (CHAR_INFO *it = pbar->bar + 0, *const end = it + len; it < end; ++it)
  {
    it->Char.AsciiChar = ' ';
    it->Attributes = col;
  }

  // The current percentage will be displayed only if the bar has a length of at least 4 character cells. (3 digits + percent sign)
  const short half = len >> 1;
  if (half > 1)
  {
    (pbar->bar + half)->Char.AsciiChar = '0';
    (pbar->bar + half + 1)->Char.AsciiChar = '%';
  }

  // Write the progress bar.
  pbar->bar_target = (SMALL_RECT){ x, y, x + len - 1, y };
  pbar->bar_metrics = pbar->bar_target;
  pbar->bar_size = (COORD){ len, 1 };
  if (!WriteConsoleOutputW(h_out, pbar->bar, pbar->bar_size, (COORD){ 0 }, &pbar->bar_metrics) || (pbar->msg = malloc(len)) == NULL)
  {
    CloseHandle(h_out);
    free(pbar);
    return NULL;
  }

  // Scroll the viewport and hide the cursor (now that we don't return early anymore).
  COORD cursor_pos = { 0, pbar->bar_metrics.Bottom };
  SetConsoleCursorPosition(h_out, cursor_pos);
  cursor_pos.Y += 2;
  SetConsoleCursorPosition(h_out, cursor_pos);
  GetConsoleCursorInfo(h_out, &pbar->cursor_inf);
  pbar->cursor_inf.bVisible = FALSE;
  SetConsoleCursorInfo(h_out, &pbar->cursor_inf);

  // Write the message.
  if (msg != NULL)
  {
    COORD pos = { pbar->bar_metrics.Left, pbar->bar_metrics.Bottom + 1 };
    const size_t msglen = strlen(msg);
    const DWORD barlen = (DWORD)len;
    DWORD processed = 0UL;
    if (msglen < barlen)
    {
      WriteConsoleOutputCharacterA(h_out, msg, (DWORD)msglen, pos, &processed);
      pos.X += (short)msglen;
      FillConsoleOutputCharacterW(h_out, L' ', (DWORD)(barlen - msglen), pos, &processed);
    }
    else
      WriteConsoleOutputCharacterA(h_out, msg, barlen, pos, &processed);
  }

  // Initialize remaining members.
  InitializeCriticalSection(&pbar->mtx);
  pbar->h_out = h_out;
  pbar->steps = steps;
  pbar->color = (WORD)(((col & 0x0F) << 4) | ((col & 0xF0) >> 4));

  return pbar;
}

bool pbar_push(progress_bar pbar, const char *msg)
{
  if (pbar == NULL)
    return false;

  EnterCriticalSection(&pbar->mtx); // Synchronize access if the function is called asynchronously.
  if (pbar->signal) // Check whether it has been signaled to get off instantly.
  {
    LeaveCriticalSection(&pbar->mtx);
    return false;
  }

  if (pbar->h_thread != NULL)
  {
    WaitForSingleObject(pbar->h_thread, 1000UL); // Wait for the previous write_bar thread.
    CloseHandle(pbar->h_thread);
    pbar->h_thread = NULL;
  }

  ++pbar->step_count;

  // Prepare the message.
  if (msg != NULL)
  {
    const size_t msglen = strlen(msg), barlen = (size_t)(pbar->bar_metrics.Right - pbar->bar_metrics.Left + 1);
    if (msglen < barlen)
    {
      memcpy(pbar->msg, msg, msglen);
      memset(pbar->msg + msglen, ' ', barlen - msglen);
    }
    else
      memcpy(pbar->msg, msg, barlen);
  }
  else
    *pbar->msg = '\0';

  // Calculate and draw the bar in a thread because writing to the console window
  // is slow and we don't want to waste time for a side task like a progress bar.
  // Don't worry whether the creation of the thread succeeded. Failing is negligibly
  // rare and the bar would still be updated properly in any subsequent execution.
  pbar->h_thread = CreateThread(NULL, 0U, write_bar, (void *)pbar, 0UL, NULL);

  // Last step might be reached here.
  if (pbar->steps == pbar->step_count)
  {
    pbar->signal = -1; // Signal other instances of this function to bail out immediately.
    LeaveCriticalSection(&pbar->mtx);
    return false;
  }

  LeaveCriticalSection(&pbar->mtx);
  return true;
}

short pbar_get_line(progress_bar pbar)
{
  if (pbar == NULL)
    return (short)-1;

  return pbar->bar_metrics.Top;
}

short pbar_get_length(progress_bar pbar)
{
  if (pbar == NULL)
    return (short)-1;

  return pbar->bar_metrics.Right - pbar->bar_metrics.Left + 1;
}

size_t pbar_get_steps(progress_bar pbar)
{
  if (pbar == NULL)
    return (size_t)-1;

  EnterCriticalSection(&pbar->mtx); // Synchronize access.
  const size_t step_count = pbar->step_count;
  LeaveCriticalSection(&pbar->mtx);

  return step_count;
}

void pbar_finalize(progress_bar *ptrpbar, const char *msg, bool show_cursor)
{
  if (ptrpbar == NULL || *ptrpbar == NULL)
    return;

  EnterCriticalSection(&(*ptrpbar)->mtx); // Synchronize access.
  (*ptrpbar)->signal = -1; // Signal instances of pbar_push to bail out immediately.
  LeaveCriticalSection(&(*ptrpbar)->mtx);

  if ((*ptrpbar)->h_thread != NULL)
  {
    WaitForSingleObject((*ptrpbar)->h_thread, 1000UL); // Wait for the write_bar thread.
    CloseHandle((*ptrpbar)->h_thread);
    (*ptrpbar)->h_thread = NULL;
  }

  DWORD processed = 0UL;

  // Write the final message.
  if (msg != NULL)
  {
    const size_t msglen = strlen(msg);
    const DWORD barlen = (DWORD)((*ptrpbar)->bar_metrics.Right - (*ptrpbar)->bar_metrics.Left + 1);
    COORD pos = { (*ptrpbar)->bar_metrics.Left, (*ptrpbar)->bar_metrics.Bottom + 1 };
    if (msglen < barlen)
    {
      WriteConsoleOutputCharacterA((*ptrpbar)->h_out, msg, (DWORD)msglen, pos, &processed);
      pos.X += (short)msglen;
      FillConsoleOutputCharacterW((*ptrpbar)->h_out, L' ', (DWORD)(barlen - msglen), pos, &processed);
    }
    else
      WriteConsoleOutputCharacterA((*ptrpbar)->h_out, msg, barlen, pos, &processed);
  }

  const COORD cursor_pos = { 0, (*ptrpbar)->bar_metrics.Bottom + 2 };
  SetConsoleCursorPosition((*ptrpbar)->h_out, cursor_pos); // Move the cursor below the message.
  if (show_cursor)
  {
    (*ptrpbar)->cursor_inf.bVisible = TRUE;
    SetConsoleCursorInfo((*ptrpbar)->h_out, &(*ptrpbar)->cursor_inf); // Make the cursor visible.
  }

  // Work around the background color of the WSL prompt not being updated properly.
  WORD attr = 0;
  if (ReadConsoleOutputAttribute((*ptrpbar)->h_out, &attr, 1UL, cursor_pos, &processed))
    WriteConsoleOutputAttribute((*ptrpbar)->h_out, &attr, 1UL, cursor_pos, &processed);

  // Clean up.
  CloseHandle((*ptrpbar)->h_out);
  DeleteCriticalSection(&(*ptrpbar)->mtx);
  free((*ptrpbar)->msg);
  free((*ptrpbar));
  *ptrpbar = NULL;
}
