/*
Copyright (c) 2020 Steffen Illhardt

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
the Software, and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#ifndef CONPROGRESSBAR_H_INCLUDED__
#define CONPROGRESSBAR_H_INCLUDED__

#ifndef __cplusplus
# include <stdbool.h>
# include <stddef.h>
#else
# include <cstddef>
  extern "C" {
#endif

/** \brief  Handle of an opaque structure holding information about the status
 *           of the progress bar. */
typedef struct tag_progress_bar *progress_bar;

/** \brief  Create and draw the empty progress bar using specified metrics,
 *           hide the cursor.
 *          The returned progress_bar handle has to be released using
 *           pbar_finalize.
 * \param  steps  Total number of steps that represent 100%. (must not be 0)
 *                Make sure steps is exactly the number of the following calls
 *                 of function pbar_push.
 * \param  x      Alignment of the progress bar in horizontal direction.
 *                0 is the left edge of the console screen buffer.
 *                Any negative value leads to positioning at 1.
 *                The function fails if this position is less than one column
 *                 away from the right edge of the console screen buffer.
 * \param  y      Alignment of the progress bar in vertical direction.
 *                0 is the position at the top edge.
 *                Any negative value leads to positioning at the beginning of
 *                 the next empty line based on the current cursor position,
 *                 furthermore the "abs" parameter is ignored.
 *                The function fails if this position is less than 3 lines
 *                 above the bottom edge of the console screen buffer.
 * \param  abs    If true, the y alignment is based on the top of the unscrolled
 *                 console screen buffer.
 *                If false, the y alignment is based on the top of the viewport
 *                 (currently visible part of the screen buffer).
 *                This parameter is ignored if the "y" parameter is negative.
 * \param  len    Length of the progress bar in characters. (must not be 0)
 *                It gets automatically shortened to fit if it would exceed the
 *                 line.
 *                Any negative value makes the progress bar end one character
 *                 before the right edge of the console screen buffer.
 * \param  col    Byte representing the colors of the progress bar.
 *                The high 4 bits represent the empty part of the bar, the low
 *                 4 bits represent the progress part of the bar.
 *                For the sake of ease it's recommended to specify the color as
 *                 a hexadecimal number of 2 hex digits having the following
 *                 meaning each:
 *                 | 0 Black | 4 Red    | 8 Gray         | C Bright Red    |
 *                 | 1 Blue  | 5 Purple | 9 Bright Blue  | D Bright Purple |
 *                 | 2 Green | 6 Yellow | A Bright Green | E Bright Yellow |
 *                 | 3 Aqua  | 7 White  | B Bright Aqua  | F Bright White  |
 *                Example: 0xF1 is bright white for the empty part and blue
 *                 for the progress part of the bar.
 * \param  msg    Null-terminated message, written below the bar line.
 *                If NULL is passed, the message line is left unchanged.
 * \return  A progress_bar handle. NULL if pbar_create failed. */
progress_bar pbar_create(size_t steps, short x, short y, bool abs, short len, unsigned char col, const char *msg);

/** \brief  Advance the internal step counter for a finalized item, draw the
 *           updated progress bar.
 *          This function is thread-safe and might be called concurrently.
 * \param  pbar  The progress_bar handle.
 * \param  msg   Null-terminated message, written below the bar line.
 *               If NULL is passed, the message line is left unchanged.
 * \return  true  as long as the internal counter is less than the number of
 *           steps (see pbar_create).
 *          false  if the internal counter reached the number of steps and the
 *           progress bar is at 100%. */
bool pbar_push(progress_bar pbar, const char *msg);

/** \brief  Get the absolute line index of the progress bar.
 * \param  pbar  The progress_bar handle.
 * \param  msg   Null-terminated message, written below the bar line.
 *               If NULL is passed, the message line is left unchanged.
 * \return  Position of the bar line in y direction, based on the top of the
 *           console screen buffer.
 *          Returns -1 if a NULL pointer is passed to the pbar parameter. */
short pbar_get_line(progress_bar pbar);

/** \brief  Get the length of the progress bar.
 * \param  pbar  The progress_bar handle.
 * \return  Length of the progress bar.
 *          Returns -1 if a NULL pointer is passed to the pbar parameter. */
short pbar_get_length(progress_bar pbar);

/** \brief  Get the current number of steps processed.
 * \param  pbar  The progress_bar handle.
 * \return  Value of the internal step counter.
 *          Returns (size_t)-1 if NULL is passed to the pbar parameter. */
size_t pbar_get_steps(progress_bar pbar);

/** \brief  Free allocated memory, move the cursor to the next line below the
*            message.
 *          The progress_bar handle is invalidated and set to NULL.
 *          Optionally, the cursor can be made visible.
 *          The pbar_finalize function can also be called to abort the progress
 *           bar at any time.
 *          Must not be called concurrently.
 * \param  ptrpbar      Pointer to the progress_bar handle.
 * \param  msg          Null-terminated message, written below the bar line.
 *                      If NULL is passed, the message line is left unchanged.
 * \param  show_cursor  If true, the cursor is visible again. */
void pbar_finalize(progress_bar *ptrpbar, const char *msg, bool show_cursor);

#ifdef __cplusplus
}
#endif

#endif // CONPROGRESSBAR_H_INCLUDED__
